using System;
using Xunit;

namespace GradeBook.Tests
{
    using GradeBook;
 
    public delegate string WriteLogDelegate(string logMessage);

    public class TypeTests
    {
        int count = 0;

        [Fact]
        public void WriteLogDelegateCanPointToMethod()
        {
            string msg = "test";

            WriteLogDelegate log = new WriteLogDelegate(ReturnMessage); // Instantiate a delegate
            log += ReturnMessage2; // Multi-cast delegate
            var result = log(msg); // Invoking a delegate
            Assert.Equal(result, msg);
            Assert.Equal(count, 2);
        }

        string ReturnMessage(string message)
        {
            count += 1;
            return message;
        }
        string ReturnMessage2(string message)
        {
            count += 1;
            return message;
        }

        [Fact]
        public void ValueTypesAlsoPassByValue()
        {
            var x = GetInt();
            Assert.Equal(3,x);

            SetInt(ref x);
            Assert.Equal(42, x);
        }

        private void SetInt(ref int x)
        {
            x = 42;
        }

        private int GetInt()
        {
            return 3;
        }

        [Fact]
        public void StringsBehaveLikeValueTypes()
        {
            string name = "Jason";
            MakeUppserCase(name); // try to change String like a reference

            Assert.Equal("Jason", name); // ToUpper does not work on String, although it's reference. It's copied as value
        }

        private void MakeUppserCase(string arg)
        {
            arg.ToUpper();
        }

        [Fact]
        public void CSharpCanPassByReference()
        {
            var book1 = GetBook("Book1");
            GetBookSetName(ref book1, "New Name");

            Assert.Equal(book1.Name, "New Name");
        }

        private void GetBookSetName(ref InMemoryBook book, string name)
        {
            book = new InMemoryBook(name);
            book.Name = name;
        }

        [Fact]
        public void CSharpIsPassByValue()
        {
            var book1 = GetBook("Book1");
            GetBookSetName(book1, "New Name");

            Assert.Equal(book1.Name, "Book1");
        }

        private void GetBookSetName(InMemoryBook book, string name)
        {
            book = new InMemoryBook(name);
            book.Name = name;
        }

        [Fact]
        public void CanSetNameFromReference()
        {
            var book1 = GetBook("Book1");
            SetName(book1, "New Name");

            Assert.Equal(book1.Name, "New Name");
        }

        private void SetName(InMemoryBook book, string name)
        {
            book.Name = name;
        }

        [Fact]
        public void GetBookReturnsDifferentObjects()
        {
            var book1 = GetBook("Book1");
            var book2 = GetBook("Book2");

            Assert.Equal("Book1", book1.Name);
            Assert.Equal("Book2", book2.Name);
        }

        [Fact]
        public void TwoVarsCanReferenceSameObject()
        {
            var book1 = GetBook("Book1");
            var book2 = book1;

            Assert.Equal(book2.Name, book1.Name);
            Assert.True(Object.ReferenceEquals(book1, book2));
        }


        private InMemoryBook GetBook(string name)
        {
            return new InMemoryBook(name);
        }
    }
}
